/*
** Module   : RadioSet.java
** Abstract : RadioSet implementation for CHUI Radio-set
**
** Copyright (c) 2005-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description------------------------------
** 001 EVL 20050301   @20143 Creation based on CHARVA JPanel class.
**                           Necessary modification to get compiled
** 002 EVL 20050303   @20149 Adding initial data arbitrary association
**                           for radio buttons
** 003 EVL 20050304   @20170 Handling of auto-resize feature
** 004 EVL 20050524   @21284 Adding the Progress like methods
** 005 EVL 20050604   @21405 Changing the inner variable names to sync
**                           with the rest of the project.
** 006 EVL 20050604   @21406 Adding the Progress specific methods to work
**                           with Progress radio-set attributes
** 007 EVL 20050607   @21435 Syntax for comments checking
** 008 EVL 20050607   @21438 Removing the visible member. It is completely
**                           duplicated in CHARVA's Component superclass
**                           so we do not need to double this work. Also 
**                           the methods to work with the HIDDEN attribute
**                           are modified to reflect this change
** 009 EVL 20050723   @21806 Properly handling of the change the
**                           RADIO_BUTTONS attribute.
** 010 EVL 20050822   @22199 Using the P2JFrame/Container interfaces to
**                           detect the class.
** 011 SIY 20050831   @22510 Refactoring (see @@22394-22409). 
** 012 SIY 20050907   @22646 Fixed use of @inheritDoc tag in javadoc.
** 013 SIY 20050909   @22668 Fixed double initialization of items.
** 014 EVL 20050912   @22676 Adding Validateable interface to this widget.
** 015 EVL 20050913   @22698 Adding focus listener interface.
** 016 EVL 20050914   @22714 Adding one more focus lost event to identify
**                           actual focus lost for radio-set.
** 017 EVL 20050916   @22736 Returned value class is changed to be the
**                           BaseDataType.
** 018 EVL 20050920   @22762 Adding the implementation of the new method
**                           isModified() for the Validateable interface.
** 019 EVL 20050921   @22766 Adding method to be able to undo the widget's
**                           value.
** 020 SIY 20050920   @22846 Added support for SENSITIVE and SELECTED 
**                           attributes.
** 021 GES 20060114   @24207 Implemented DataContainer interface and the
**                           core methods.
** 022 SIY 20060131   @24208 Use ThinClient special method instead of 
**                           direct calls to Window.message().
** 023 SIY 20060209   @24462 Major refactoring and cleanups.
** 024 SIY 20060210   @24519 Added getSelection() method, added 
**                           ControlSetComponent interface.
** 025 SIY 20060222   @24640 Added necessary methods to support changes in
**                           DataContainer interface.
** 026 NVS 20060227   @24869 Key actions KA_* are used instead of virtual
**                           keys VK_* for all editing operations.
** 027 NVS 20060314   @25069 Added reset() method that resets the set of
**                           buttons to the first button selected.
** 028 SIY 20060322   @25212 Added setNoLabels() to match new 
**                           LabeledWidget interface.
** 029 SIY 20060410   @25465 Fixed size calculation. 
** 030 SIY 20060417   @25593 More minor fixes and cleanups.
** 031 SIY 20060515   @26155 Encapsulated access to help field.
** 032 SIY 20060515   @26156 Added getColumnLabelText().
** 033 SIY 20060518   @26277 Changed initial widget state to disabled.
** 034 NVS 20060630   @27729 Reviewed requestFocus() calls. Those in
**                           processKeyEvents() are safe, and those in
**                           refreshItems() are *likely* to be safe.
**                           However, nextFocus() and previousFocus()
**                           calls made from processKeyEvent() are
**                           definitely UNSAFE and need separate
**                           consideration.
** 035 NVS 20060703   @27803 getParent.nextFocus() replaced with
**                           ThinClient.nextFocus() call to maintain
**                           Progress semantics. Similar replacement has
**                           been done to getParent.previousFocus().
** 036 SIY 20060824   @28855 Removed getP2JType() method.
** 037 NVS 20061204   @31559 Switched to using a private widgetFocused
**                           flag instead of config.selected
** 038 NVS 20061214   @31692 Added getIndex() method and enabled drawing
**                           of uninitialized radio buttons.
** 039 NVS 20061221   @31733 Currently focused button within the group
**                           should be preserved when tabbing through the
**                           widgets of the frame.
** 040 NVS 20070426   @33254 Enabled state is checked instead of SENSITIVE
**                           attribute.
** 041 SIY 20070807   @34805 Added isColumnFixed() method.
** 042 SVL 20071221   @36531 Label is shown/hidden together with the
**                           widget. 
** 043 SIY 20080320   @37600 Fixed handling of the SCREEN-VALUE and data
**                           value.
** 044 NVS 20080401   @37779 Added isFocusable() method implementation.
** 045 SIY 20080411   @37921 Significant rewrite in order to accommodate
**                           changes in focus handling.
**                           Ref: #37923
** 046 SIY 20100619          Switching to own implementation of character UI.
** 047 SIY 20100804          Changes in imports, some refactorings.
** 048 SIY 20100822          Refactored AttributeContainer support.
** 049 SIY 20101122          Switching from KeyEvent to KeyInput.
** 050 SIY 20111101          Updated imports, minor cleanups/refactorings.
** 051 SIY 20111202          Refactored and moved from CHUI package.
** 052 SIY 20111221          Switched to widget factory.
** 053 CA  20140805          Added generics to force the abstract widget classes to not use
**                           implementation-specific code: the abstract widgets will have access
**                           only to the base APIs provided by the OutputManager.
** 054 HC  20140911          GUI coordinate support - data types refactoring.
** 055 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 056 HC  20141021          Implemented window move/position support.
** 057 IAS 20150204          Implemented ADD-LAST method.
*/

package com.goldencode.p2j.ui.client;


import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * RadioSet is a implementation of radio-set widget, a mutually exclusive set
 * of {@link RadioButton} objects.
 * <p>
 * WARNING: this class needs updates to support BaseDataType instances. It
 * is currently hard coded to STRINGS.
 * <p>
 * TODO<SIY>: this class does not require dedicated CHUI/GUI versions because
 * it delegates drawing to buttons, but refreshItems() requires update in order
 * to maintain buttons physical position/size according to GUI rules.
 *  
 */
public class RadioSet<O extends OutputManager<?>>
extends LabeledBorderedPanel<O>
implements FocusListener,
           DataContainer,
           ControlSetComponent
{
   /** Component attribute container */
   private RadioSetConfig config = null;
   
   /** current button */
   private int current = 0;
   
   /** Button collection. */
   private ArrayList<RadioButton<O>> buttons = new ArrayList<>();
   
   /**
    * Special constructor used to restore component using given configuration.
    * <p>
    * Call {@link #initialize} to finish the widget initialization.
    *
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public RadioSet(WidgetId id, WidgetConfig cfg)
   {
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, WidgetConfig cfg)
   {
      setEnabled(false);
      config = (RadioSetConfig) ConfigManager.getInstance().resolveWidgetConfig(id, cfg);
      ConfigManager.getInstance().replaceConfig(config, cfg);
      addFocusListener(this);
   }
   
   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @SuppressWarnings("unchecked")
   @Override
   public RadioSetConfig config()
   {
      return config;
   }
   
   /**
    * This overriding is used to move the focus from this widget to the
    * next/previous one, rather than moving between radio buttons inside the
    * current radio-set.
    * 
    * @param   keyEvent
    *          The key event to be processed.
    */
   @Override
   public void processKeyEvent(KeyInput keyEvent)
   {
      if (keyEvent.isConsumed())
         return;
      
      int action = keyEvent.actionCode();
      int key = keyEvent.keyCode();

      int save_current = current;
      
      if (action == Keyboard.KA_CURSOR_UP || 
               action == Keyboard.KA_CURSOR_LEFT)
      {
         for (int i = current - 1; i >= 0; i--)
         {
            if (buttons.get(i).isSelectable())
            {
               current = i;
               break;
            }
         }
      }
      else if (action == Keyboard.KA_CURSOR_DOWN ||
               action == Keyboard.KA_CURSOR_RIGHT)
      {
         for (int i = current + 1; i < buttons.size(); i++)
         {
            if (buttons.get(i).isSelectable())
            {
               current = i;
               break;
            }
         }
      }
      else if (action == Keyboard.KA_TAB || 
               action == Keyboard.KA_BACK_TAB)
      {
         if (action == Keyboard.KA_TAB)
            ThinClient.nextFocus(this);
         else
            ThinClient.previousFocus(this);
      }
      else if (key == ' ')
      {
         RadioButton<O> currentButton = buttons.get(current);
         
         if (!currentButton.isSelected())
         {
            selectButton(currentButton);
            config.setModified(true);
            EventManager.postEvent(new KeyInput(this,
                  EventType.KEY_PRESSED,
                  Keyboard.SE_VALUE_CHANGED,
                  true));
        }
      }
      else
         super.processKeyEvent(keyEvent);
      
      if (current != save_current)
      {
         buttons.get(save_current).setFocused(false);
         buttons.get(current).setFocused(true);
         
         repaint();
         requestSync();
      }
   }

   /**
    * Must return the value of the radio-set as a BaseDataType.
    * 
    * @return  The BaseDataType value of the selected button inside the
    *          button group.
    */
   @Override
   public BaseDataType getValue()
   {
      RadioButton<O> button = getSelected();
      
      return (button != null) ? button.getValue() : null;
   }
   
   /**
    * Get text representation of the currently selected value.
    * 
    * @return  The screen representation of the value.
    */
   @Override
   public String getText()
   {
      RadioButton<O> button = getSelected();
      
      return (button != null) ? button.getText() : null;
   }
   
   /**
    * Sets the value of the radio-set (the selected button) based on a
    * BaseDataType value.
    * 
    * @param    value
    *           The BaseDataType value of the button to be selected inside
    *           the button group.
    */
   @Override
   public void setValue(BaseDataType value)
   {
      for (RadioButton<O> button : buttons)
      {
         if (button.getValue().equals(value))
            selectButton(button);
      }
   }

   /**
    * Select specified button and remove current selection.
    * 
    * @param   button
    *          Button to select.
    */
   private void selectButton(RadioButton<O> button)
   {
      if (button == null)
         return;
      
      RadioButton<O> selected = getSelected();
      
      if (selected != null)
         selected.setSelected(false);
      
      button.setSelected(true);
      
      repaint();
      requestSync();
   }
   
   /**
    * Resets the value of the radio-set (the selected button) to the first
    * button of the set.
    */
   public void reset()
   {
      if (buttons.isEmpty())
      {
         return;
      }
      RadioButton<O> b = buttons.get(0);
      
      if (b.isSelected())
      {
         config.setModified(true);
         setInitial();
      }
      else
         selectButton(b);
   }

   /**
    * Indicates whether this component can be traversed using TAB or BACK-TAB
    * events.
    * 
    * @return  <code>true</code> if widget can be traversed.
    */
   @Override
   public boolean focusTraversable() 
   {
       return isEnabled() && isVisible();
   }
   
   /**
    * Return reference to this, since REDIO-SET is always retains focus on
    * itself.
    * 
    * @return  this
    */
   @Override
   public Widget<O> currentFocus()
   {
      return this;
   }
   
   /**
    * Getting the modification state of the widget.
    * 
    * @return  <code>true</code> if the widget value was modified,
    *          <code>false</code> otherwise.
    */
   @Override
   public boolean isModified()
   {
      return config.modified;
   }

   /**
    * Set up the width of the button dynamically.
    * 
    * @param   newWidth
    *          The new width of the radio-button.
    */
   public void setWidthChars(double newWidth)
   {
      newWidth = Coordinate.scale(newWidth);
      
      if (newWidth <= 0 || config.horizontal)
         return;

      for (RadioButton<O> button : buttons)
         button.setWidthChars(newWidth);

      repaint();
   }

   /**
    * Getting the width of the button. The final width includes the button's
    * label size.
    * 
    * @return  The integer number of the width of the radio-buttons.
    */
   public double getCharWidth()
   {
      return config.widthChars;
   }

   /**
    * Draw RadioSet widget.
    */
   @Override
   public void draw()
   {
      for (RadioButton<O> button : buttons)
         button.draw();
   }
   
   /**
    * This is called when the radio-set takes the input focus. When the help
    * string is defined for the radio-set it should be displayed at the window
    * status line.
    *
    * @param   focusEvent
    *          The focus event to be processed.
    */
   @Override
   public void onFocusGained(FocusEvent focusEvent)
   {
      if (focusEvent == null) //Just make argument referenced
         return;
      
      buttons.get(current).setFocused(true);
      
      ThinClient.setHelp(config.help);

      repaint();
      requestSync();
   }

   /**
    * This is called when the radio-set looses the input focus.
    *
    * @param   focusEvent
    *          The focus event to be processed.
    */
   @Override
   public void onFocusLost(FocusEvent focusEvent)
   {
      if (focusEvent == null)
         return;
      
      buttons.get(current).setFocused(false);

      repaint();
      requestSync();
   }

   /**
    * Refresh widget look using new set of buttons.
    */
   @Override
   public void refreshItems()
   {
      int selectedId = -1;
      for (RadioButton button: buttons) 
      {
         if (button.isSelected())
         {
            selectedId = button.getId();
            break;
         }
      }
      buttons.clear();
      
      ControlSetItem[] items = config.items;
      
      Point pNext = new Point(0, 0);

      RadioButton<O> selected = null;
      
      Dimension dim = new Dimension(0,config.heightChars);
      WidgetFactory<O> factory = screen().getFactory();
      
      int bwidth = -1;
      if (config.horizontal && config.expand)
      {
         for (ControlSetItem item: items)
         {
            character label = item.getLabel();
            if (label != null)
            {
               bwidth = Math.max(bwidth, label.toStringMessage().length());
            }
         }
         bwidth += 3; // "( )".length()
      }
      
      int maxSize = (int)(config.horizontal ? config.widthChars : config.heightChars);
      int size = 0;
      for (int i = 0; i < items.length; i++)
      {
           
         RadioButton<O> button = factory.createRadioButton(config, 
                                                           items[i].getLabel().getValue(), 
                                                           items[i].getValue());
         
         if (bwidth >= 0)
            button.setWidthChars(bwidth);
         
         size += config.horizontal ? button.width() : button.height();
         if (size > maxSize)
         {
            break;
         }

         button.setSelectable(items[i].isSelectable());
         button.setId(items[i].getId());
         
         button.setParent(this);
         button.setLocation(pNext);

         buttons.add(button);

         if (config.horizontal)
         {
            size++;
            pNext.x += button.width() + 1;
         }
         else
            pNext.y += button.height();
         
         if (items[i].getId() == selectedId)
         {
            selected = button;
            config.initial = i;
         }

      }
      
      if (selected != null)
      {
         selected.setSelected(true);
      }
      else if (buttons.size() > 0) 
      {
         for (int i = 0; i < items.length; i++)
         {
            selected = buttons.get(0);
            config.initial = 0;
            selected.setSelected(true);
            if (buttons.get(i).isSelectable())
            {
               selected = buttons.get(i);
               config.initial = i;
               selected.setSelected(true);
               break;
            }
            current = config.initial;
         }
      }
      
      setSize(new Dimension(config.widthChars,config.heightChars));

      repaint();
   }

   /**
    * Request indexes of selected items.
    * 
    * @return  array of 0-based indexes of selected items.
    */
   @Override
   public int[] getSelection()
   {
      int index = buttons.indexOf(getSelected());
      
      if (index < 0)
         index = 0;

      return new int[] {index};
   }
   
   /**
    * Request index of some button.
    * 
    * @param   rb 
    *          Radio button.
    *          
    * @return  0-based index of the button or -1 if not in the set.
    */
   public int getIndex(RadioButton<O> rb)
   {
      return buttons.indexOf(rb);
   }
   
   /**
    * Track index of currently selected button. 
    */
   private void setInitial()
   {
      config.initial = getSelection()[0];
   }
   
   /**
    * Get selected button.
    * 
    * @return  Selected button.
    */
   public RadioButton<O> getSelected()
   {
      for (RadioButton<O> button : buttons)
      {
         if (button.isSelected())
            return button;
      }
      
      return null;
   }
}
